# encoding: utf-8
"""
Represents a connection to the ECS service.
"""

import warnings

import six
import time
import json
import base64
from footmark.ecs.config import *
from footmark.connection import ACSQueryConnection
from footmark.ecs.zone import Zone
from footmark.ecs.instance_type import InstanceType, InstanceTypeFamily
from footmark.ecs.regioninfo import RegionInfo
from footmark.ecs.instance import Instance
from footmark.ecs.image import Image
from footmark.ecs.securitygroup import SecurityGroup
from footmark.ecs.volume import Disk
from footmark.ecs.networkinterface import NetworkInterfaceSet
from footmark.exception import ECSResponseError
from footmark.resultset import ResultSet
from aliyunsdkcore.acs_exception.exceptions import ServerException
# from aliyunsdkecs.request.v20140526.CreateInstanceRequest import Request import
# from aliyunsdkcore.auth.composer.rpc_signature_composer import ServerException


class ECSConnection(ACSQueryConnection):
    SDKVersion = '2014-05-26'
    DefaultRegionId = 'cn-hangzhou'
    DefaultRegionName = u'杭州'.encode("UTF-8")
    ResponseError = ECSResponseError

    def __init__(self, acs_access_key_id=None, acs_secret_access_key=None,
                 region=None, sdk_version=None, security_token=None, user_agent=None):
        """
        Init method to create a new connection to ECS.
        """
        if not region:
            # region = RegionInfo(self, self.DefaultRegionName,
            #                     self.DefaultRegionId)
            region = self.DefaultRegionId
        self.region = region
        if sdk_version:
            self.SDKVersion = sdk_version

        self.ECSSDK = 'aliyunsdkecs.request.v' + self.SDKVersion.replace('-', '')

        super(ECSConnection, self).__init__(acs_access_key_id=acs_access_key_id,
                                            acs_secret_access_key=acs_secret_access_key,
                                            region=self.region, product=self.ECSSDK,
                                            security_token=security_token, user_agent=user_agent)

    def build_filter_params(self, params, filters):
        if not isinstance(filters, dict):
            return

        flag = 1
        for key, value in filters.items():
            acs_key = key
            if acs_key.startswith('tag:'):
                while ('set_Tag%dKey' % flag) in params:
                    flag += 1
                if flag < 6:
                    params['set_Tag%dKey' % flag] = acs_key[4:]
                    params['set_Tag%dValue' % flag] = filters[acs_key]
                flag += 1
                continue
            if key == 'group_id':
                if not value.startswith('sg-') or len(value) != 12:
                    warnings.warn("The group-id filter now requires a security group "
                                  "identifier (sg-*) instead of a security group ID. "
                                  "The group-id " + value + "may be invalid.",
                                  UserWarning)
                params['set_SecurityGroupId'] = value
                continue
            if not isinstance(value, dict):
                acs_key = ''.join(s.capitalize() for s in acs_key.split('_'))
                params['set_' + acs_key] = value
                continue

            self.build_filters_params(params, value)

    def build_tags_params(self, params, tags, max_tag_number=None):
        tag_no = 1
        if tags:
            for key, value in tags.items():
                if tag_no > max_tag_number:
                    break
                if key:
                    self.build_list_params(params, key, 'Tag' + str(tag_no) + 'Key')
                    self.build_list_params(params, value, 'Tag' + str(tag_no) + 'Value')
                    tag_no += 1

    def create_instances(self, **kwargs):
        """
        create an instance in ecs

        :type image_id: string
        :param image_id: ID of an image file, indicating an image selected
            when an instance is started

        :type instance_type: string
        :param instance_type: Type of the instance

        :type group_id: string
        :param group_id: ID of the security group to which a newly created
            instance belongs

        :type zone_id: string
        :param zone_id: ID of a zone to which an instance belongs. If it is
            null, a zone is selected by the system

        :type instance_name: string
        :param instance_name: Display name of the instance, which is a string
            of 2 to 128 Chinese or English characters. It must begin with an
            uppercase/lowercase letter or a Chinese character and can contain
            numerals, “.”, “_“, or “-“.

        :type description: string
        :param description: Description of the instance, which is a string of
            2 to 256 characters.

        :type internet_data: list
        :param internet_data: It includes Internet charge type which can be
            PayByTraffic or PayByBandwidth, max_bandwidth_in and max_bandwidth_out

        :type host_name: string
        :param host_name: Host name of the ECS, which is a string of at least
            two characters. “hostname” cannot start or end with “.” or “-“.
            In addition, two or more consecutive “.” or “-“ symbols are not
            allowed.

        :type password: string
        :param password: Password to an instance is a string of 8 to 30
            characters

        :type io_optimized: string
        :param io_optimized: values are (1) none: none I/O Optimized
            (2) optimized: I/O Optimized

        :type system_disk: list
        :param system_disk: It includes disk_category, disk_size,
            disk_name and disk_description

        :type disks: list
        :param disks: It includes device_category, device_size,
            device_name, device_description, delete_on_termination
            and snapshot

        :type vswitch_id: string
        :param vswitch_id: When launching an instance in VPC, the
            virtual switch ID must be specified

        :type private_ip: string
        :param private_ip: Private IP address of the instance, which cannot be specified separately.

        :type count: integer
        :param count: Create No. of Instances

        :type allocate_public_ip: bool
        :param allocate_public_ip: Allocate Public IP Address to Instance

        :type bind_eip: string
        :param bind_eip: Bind Elastic IP Address

        :type instance_charge_type: string
        :param instance_charge_type: instance charge type

        :type: period: integer
        :param period: The time that you have bought the resource,
            in month. Only valid when InstanceChargeType is set as
            PrePaid. Value range: 1 to 12

        :type: auto_renew: bool
        :param auto_renew: Whether automatic renewal is supported.
            Only valid when InstanceChargeType is set PrePaid. Value
            range True: indicates to automatically renew
                  False，indicates not to automatically renew
            Default value: False.

        :type: auto_renew_period: int
        :param auto_renew_period: The period of each automatic renewal. Required when AutoRenew is True.
        The value must be the same as the period of the created instance.

        :type: ids: list
        :param ids: A list of identifier for this instance or set of
            instances, so that the module will be idempotent with
            respect to ECS instances.

        :type instance_tags: list
        :param instance_tags: A list of hash/dictionaries of instance
            tags, '[{tag_key:"value", tag_value:"value"}]', tag_key
            must be not null when tag_value isn't null

        :type wait: string
        :param wait: after execution of method whether it has to wait for some time interval

        :type wait_timeout: int
        :param wait_timeout: time interval of waiting

        :rtype: dict
        :return: Returns a dictionary of instance information about
            the instances started/stopped. If the instance was not
            able to change state, "changed" will be set to False.
            Note that if instance_ids and instance_tags are both non-
            empty, this method will process the intersection of the two

        """

        instance_ids = []
        count = kwargs["count"]
        kwargs = self.format_request_kwargs(**kwargs)
        if not count:
            count = 1
        for i in range(count):
            # CreateInstance method call, returns newly created instanceId
            client_token = kwargs["client_token"]
            if count > 1:
                client_token = "{0}-{1}".format(i, client_token)
            if len(client_token) > 64:
                client_token = client_token[0:64]
            kwargs["client_token"] = client_token
            instance_ids.append(self.get_object_new(self.build_request_params(kwargs), ResultSet).instance_id)

        transition = [].extend(instance_ids)
        while transition:
            for inst in self.describe_instances(instance_ids=transition, page_size=100):
                if str(inst.status).lower() == 'stopped':
                    transition.remove(inst.id)
        # There need sleep 10s to wait instance to finish initiated.
        if instance_ids:
            time.sleep(10)

        return self.describe_instances(instance_ids=instance_ids, page_size=100)

    # Instance methods
    def get_all_instances(self, **kwargs):
    # def get_all_instances(self, zone_id=None, instance_ids=None, instance_name=None, instance_tags=None,
    #                       vpc_id=None, vswitch_id=None, instance_type=None, instance_type_family=None,
    #                       instance_network_type=None, private_ip_addresses=None, inner_ip_addresses=None,
    #                       public_ip_addresses=None, security_group_id=None, instance_charge_type=None,
    #                       spot_strategy=None, internet_charge_type=None, image_id=None, status=None,
    #                       io_optimized=None, pagenumber=None, pagesize=100):
        """
        Retrieve all the instance associated with your account. 

        :rtype: list
        :return: A list of  :class:`footmark.ecs.instance`

        """
        instances = []
        params = {}
        for key, value in kwargs.items():
            if key == "instance_ids":
                value = "[" + str(",").join(value) + "]"
                kwargs[key] = value
            if key == "tags":
                tags = []
                if isinstance(value, dict):
                    for k, v in value.items():
                        tags.append({"Key": k, "Value": v})
                kwargs[key] = value
            kwargs['Action'] = 'DescribeInstances'

            for inst in self.get_list_new(self.build_request_params(filters), ['Instances', Instance]):
                filters = {}
                filters['instance_id'] = inst.id
                volumes = self.get_all_volumes(filters)
                setattr(inst, 'block_device_mapping', volumes)
                # if inst.security_group_ids:
                #     group_ids = []
                #     security_groups = []
                #     for sg_id in inst.security_group_ids['security_group_id']:
                #         group_ids.append(str(sg_id))
                #         security_groups.append(self.get_security_group_attribute(sg_id))
                #     setattr(inst, 'security_groups', security_groups)
                instances.append(inst)
        return instances

    def describe_instances(self, **kwargs):
        """
        Retrieve all the instance associated with your account.

        :rtype: list
        :return: A list of  :class:`footmark.ecs.instance`

        """
        instances = []
        for inst in self.get_list_new(self.build_request_params(self.format_request_kwargs(**kwargs)), ['Instances', Instance]):
            volumes = self.describe_disks(instance_id=inst.id)
            setattr(inst, 'block_device_mapping', volumes)
            setattr(inst, 'user_data', base64.b64decode(self.get_object_new(self.build_request_params({"instance_id": inst.id, "Action": "DescribeUserData"}), ResultSet).user_data))
            instances.append(inst)
        return instances

    def start_instances(self, **kwargs):
        """
        Start the instances specified

        :type instance_ids: list
        :param instance_ids: A list of strings of the Instance IDs to start

        :rtype: list
        :return: A list of the instances started
        """
        instance_ids = kwargs["instance_ids"]
        kwargs = self.format_request_kwargs(**kwargs)
        if instance_ids:
            if isinstance(instance_ids, six.string_types):
                instance_ids = [instance_ids]
            for instance_id in instance_ids:
                kwargs["instance_id"]= instance_id
                try:
                    self.get_status_new(self.build_request_params(kwargs))
                except ServerException as e:
                    if e.error_code == "IncorrectInstanceStatus":
                        target = self.describe_instances(instance_ids=[instance_id])
                        if target and str(target[0].status).lower() == "running":
                            continue
                    raise e
            while instance_ids:
                for inst in self.describe_instances(instance_ids=instance_ids, page_size=100):
                    if str(inst.status).lower() == "running":
                        instance_ids.remove(inst.id)
            return True
        return False

    def stop_instances(self, **kwargs):
        """
        Stop the instances specified

        :type instance_ids: list
        :param instance_ids: A list of strings of the Instance IDs to stop

        :type force: bool
        :param force: Forces the instance to stop

        :rtype: list
        :return: A list of the instances stopped
        """
        instance_ids = kwargs["instance_ids"]
        kwargs = self.format_request_kwargs(**kwargs)
        if instance_ids:
            if isinstance(instance_ids, six.string_types):
                instance_ids = [instance_ids]
            for instance_id in instance_ids:
                kwargs["instance_id"]= instance_id
                try:
                    self.get_status_new(self.build_request_params(kwargs))
                except ServerException as e:
                    if e.error_code == "IncorrectInstanceStatus":
                        target = self.describe_instances(instance_ids=[instance_id])
                        if target and str(target[0].status).lower() == "stopped":
                            continue
                    raise e
            while instance_ids:
                for inst in self.describe_instances(instance_ids=instance_ids, page_size=100):
                    if str(inst.status).lower() == "stopped":
                        instance_ids.remove(inst.id)
            return True
        return False

    def reboot_instances(self, **kwargs):
        """
        Reboot the specified instances.

        :type instance_ids: list
        :param instance_ids: The instances to terminate and reboot

        :type force: bool
        :param force: Forces the instance to stop

        """
        instance_ids = kwargs["instance_ids"]
        kwargs = self.format_request_kwargs(**kwargs)
        if instance_ids:
            if isinstance(instance_ids, six.string_types):
                instance_ids = [instance_ids]
            for instance_id in instance_ids:
                kwargs["instance_id"]= instance_id
                try:
                    self.get_status_new(self.build_request_params(kwargs))
                except ServerException as e:
                    if e.error_code == "IncorrectInstanceStatus":
                        target = self.describe_instances(instance_ids=[instance_id])
                        if target and str(target[0].status).lower() == "stopped":
                            self.start_instances(instance_ids=[instance_id])
                    raise e
            while instance_ids:
                for inst in self.describe_instances(instance_ids=instance_ids, page_size=100):
                    status = str(inst.status).lower()
                    if status == "running":
                        instance_ids.remove(inst.id)
                    if status == "stopped":
                        self.start_instances(instance_ids=[inst.id])
            return True
        return False

    def delete_instances(self, **kwargs):
        """
        Delete the instances specified

        :type instance_ids: list
        :param instance_ids: A list of strings of the Instance IDs to terminate

        :type force: bool
        :param force: Forces the instance to stop

        :rtype: list
        :return: A list of the instance_ids terminated
        """
        instance_ids = kwargs["instance_ids"]
        kwargs = self.format_request_kwargs(**kwargs)
        if instance_ids:
            if isinstance(instance_ids, six.string_types):
                instance_ids = [instance_ids]
            for instance_id in instance_ids:
                kwargs["instance_id"]= instance_id
                self.get_status_new(self.build_request_params(kwargs))
            while instance_ids:
                if self.describe_instances(instance_ids=instance_ids):
                    continue
                break
            return True
        return False

    def allocate_public_ip_address(self, **kwargs):
        timeout = 30
        while timeout:
            try:
                return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))
            except ServerException as e:
                if e.error_code == "IncorrectInstanceStatus":
                    time.sleep(3)
                    timeout -= 3
        return False

    def describe_instance_types(self, instance_type_family=None):
        """
        Retrieve all the instance types associated with your account.

        :type instance_type_family: str
        :param instance_type_family: Family name of instance type

        :rtype: list
        :return: A list of  :class:`footmark.ecs.instance_type`

        """
        params = {}

        if instance_type_family:
            self.build_list_params(params, instance_type_family, 'InstanceTypeFamily')
        return self.get_list('DescribeInstanceTypes', params, ['InstanceTypes', InstanceType])

    def describe_zones(self, zone_id=None):
        """
            Retrieve all zones in the region.

            :type zone_id: str
            :param zone_id: Filter the zone which id is equal to zone_id

            :rtype: list
            :return: A list of  :class:`footmark.ecs.zone`

        """
        params = {}
        self.build_list_params(params, self.region, 'RegionId')
        zones = self.get_list('DescribeZones', params, ['Zones', Zone])
        if zone_id:
            zones = [zone for zone in zones if zone.id == zone_id]
        return zones

    def describe_instance_type_families(self, generation=None):
        """
            Retrieve all the instance type families associated with your account.

            :type generation: str
            :param generation: Filter the families by generation

            :rtype: list
            :return: A list of  :class:`footmark.ecs.instance_type_family`

        """
        params = {}
        self.build_list_params(params, self.region, 'RegionId')
        if generation:
            self.build_list_params(params, generation, "Generation")
        return self.get_list('DescribeInstanceTypeFamilies', params, ['InstanceTypeFamilies', InstanceTypeFamily])

    def describe_disks(self, **kwargs):
        return self.get_list_new(self.build_request_params(self.format_request_kwargs(**kwargs)), ['Disks', Disk])

    def get_all_volumes(self, zone_id=None, volume_ids=None, volume_name=None, filters=None):
        """
        Get all Volumes associated with the current credentials.

        :type volume_ids: list
        :param volume_ids: Optional list of volume ids.  If this list
                           is present, only the volumes associated with
                           these volume ids will be returned.

        :type filters: dict
        :param filters: Optional filters that can be used to limit
                        the results returned.  Filters are provided
                        in the form of a dictionary consisting of
                        filter names as the key and filter values
                        as the value.  The set of allowable filter
                        names/values is dependent on the request
                        being performed.  Check the ECS API guide
                        for details.

        :type dry_run: bool
        :param dry_run: Set to True if the operation should not actually run.

        :rtype: list of Volume
        :return: The requested Volume objects
        """
        params = {}
        if zone_id:
            self.build_list_params(params, zone_id, 'ZoneId')
        if volume_ids:
            ids = "["
            for id in volume_ids:
                ids += "\"" + str(id) + "\""
            ids += "]"
            self.build_list_params(params, ids, 'DiskIds')
        if volume_name:
            self.build_list_params(params, volume_name, 'DiskName')
        if filters:
            self.build_filter_params(params, filters)
        return self.get_list('DescribeDisks', params, ['Disks', Disk])

    def attach_key_pair(self, **kwargs):
        """
        Attach a key pair to a ecs instance

        :type instance_ids: list
        :param instance_ids: List of the instance ids

        :type key_pair_name: str
        :param key_pair_name: Name of Key Pair which is used to attach

        :return:
        """
        return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))

    def detach_key_pair(self, **kwargs):
        """
        Attach a key pair to a ecs instance

        :type instance_ids: list
        :param instance_ids: List of the instance ids

        :type key_pair_name: str
        :param key_pair_name: Name of Key Pair which is used to be detached

        :return:
        """
        return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))

    def modify_instance_attribute(self, **kwargs):
        """
        modify the instance attributes such as name, description, password and host_name

        :type instance_ids: list
        :param instance_ids: The list of Instance ID
        :type name: str
        :param name: Instance Name
        :type description: str
        :param description: Instance Description
        :type host_name: str
        :param host_name: Instance Host Name
        :type password: str
        :param password: Instance Password
        :return: A list of the instance_ids modified
        """
        return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))

    def get_instance_status(self, zone_id=None, pagenumber=None, pagesize=None):
        """
        Get status of instance

        :type zone_id: string
        :param zone_id: Optional parameter. ID of the zone to which an instance belongs

        :type pagenumber: integer
        :param pagenumber: Page number of the instance status list. The start value is 1. The default value is 1

        :type pagesize: integer
        :param pagesize: Sets the number of lines per page for queries per page. The maximum value is 50.
        The default value is 10

        :rtype: json
        :return: The result of passed instances
        """

        params = {}
        results = []

        if zone_id:
            self.build_list_params(params, zone_id, 'ZoneId')
        if pagenumber:
            self.build_list_params(params, pagenumber, 'PageNumber')
        if pagesize:
            self.build_list_params(params, pagesize, 'PageSize')

        try:
            results = self.get_object('DescribeInstanceStatus', params, ResultSet)
        except ServerException as e:
            results.append({"Error Code": e.error_code, "Error Message": e.message,
                            "RequestId": e.request_id, "Http Status": e.http_status})
        except Exception as e:
            results.append({"Error": e})

        return False, results

    def join_security_group(self, **kwargs):
        """
        Assign an existing instance to a pre existing security group

        :type instance_id: str
        :param instance_id: The instance id which are to be assigned to the security group

        :type group_id: dict
        :param group_id: ID of the security group to which a instance is to be added

        :return: Success message, confirming joining security group or error message if any
        """
        return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))

    def leave_security_group(self, **kwargs):
        """
        Remove an existing instance from given security group

        :type instance_id: str
        :param instance_id: The instance id which are to be assigned to the security group

        :type group_id: dict
        :param group_id: ID of the security group to which a instance is to be added

        :return: Success message, confirming joining security group or error message if any
        """
        return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))

    # def create_security_group(self, **kwargs):
    def create_security_group(self, **kwargs):
        """
        create and authorize security group in ecs

        :type group_name: string
        :param group_name: Name of the security group

        :type description: string
        :param description: Description of the security group

        :type group_tags: list
        :param group_tags: A list of hash/dictionaries of disk
            tags, '[{tag_key:"value", tag_value:"value"}]', tag_key
            must be not null when tag_value isn't null

        :type vpc_id: string
        :param vpc_id: The ID of the VPC to which the security group belongs

        :rtype: dict
        :return: Returns a dictionary of group information about
            the the group created/authorized. If the group was not
            created and authorized, "changed" will be set to False.
        """
        response = self.get_object_new(self.build_request_params(self.format_request_kwargs(**kwargs)), ResultSet)
        if response:
            return self.describe_security_group_attribute(security_group_id=response.security_group_id)

        return None

    def modify_security_group_attribute(self, **kwargs):
        return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))

    def authorize_security_group(self, **kwargs):
        return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))

    def authorize_security_group_egress(self, **kwargs):
        return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))

    def revoke_security_group(self, **kwargs):
        return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))

    def revoke_security_group_egress(self, **kwargs):
        return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))

    def describe_security_group_attribute(self, **kwargs):
        """
        Querying Security Group List returns the basic information about all
              user-defined security groups.

        :type  group_id: String
        :param group_id: ID of security groups id

        :type nic_type: String
        :param nic_type: Network type of security group. The choice value is 'internet' or 'intranet'.

        :type direction: String
        :param direction: The direction of security group rule. The choice value is 'egress', 'ingress' or 'all', and 'all' is default.

        :rtype: dict
        :return: Returns a dictionary of security group information

        """
        try:
            return self.get_object_new(self.build_request_params(self.format_request_kwargs(**kwargs)), SecurityGroup)
        except ServerException as e:
            if str(e.error_code) == "InvalidSecurityGroupId.NotFound":
                return None
            raise e

    def describe_security_groups(self, **kwargs):
        groups = []
        for group in self.get_list_new(self.build_request_params(self.format_request_kwargs(**kwargs)), ['SecurityGroups', SecurityGroup]):
            groups.append(self.describe_security_group_attribute(security_group_id=group.id))
        return groups

    def delete_security_group(self, **kwargs):
        """
        Delete Security Group , delete security group inside particular region.

        :type  group_id: str
        :param group_id: The Security Group ID

        :rtype: bool
        :return: A method return result of after successfully deletion of security group
        """
        # Call DescribeSecurityGroups method to get response for all running instances
        kwargs = self.format_request_kwargs(**kwargs)
        delay = 5
        timeout = DefaultTimeOut
        while True:
            try:
                self.get_status_new(self.build_request_params(kwargs))
                group = self.describe_security_group_attribute(**kwargs)
                if not group or not group.id:
                    return True
            except ServerException as e:
                if str(e.error_code) == "InvalidSecurityGroupId.NotFound":
                    return True
                elif str(e.error_code) == "DependencyViolation":
                    pass

            time.sleep(delay)
            timeout -= delay
            if timeout <= 0:
                raise Exception("Timeout: Waiting for deleting Security Group {0}, time-consuming {1} seconds. "
                                "Error: {2}".format(kwargs["security_group_id"], DefaultTimeOut-timeout, e))

    def create_disk(self, **kwargs):
        """
        create an disk in ecs

        :type zone_id: string
        :param zone_id: ID of a zone to which an instance belongs.

        :type disk_name: string
        :param disk_name: Display name of the disk, which is a string
            of 2 to 128 Chinese or English characters.

        :type description: string
        :param description: Description of the disk, which is a string of
            2 to 256 characters.

        :type disk_category: string
        :param disk_category: Displays category of the data disk
                Optional values are:
                Cloud - general cloud disk
                cloud_efficiency - efficiency cloud disk
                cloud_ssd - cloud SSD
                Default value:cloud

        :type size: integer
        :param size: Size of the system disk, in GB, values range:
                Cloud - 5 ~ 2000
                cloud_efficiency - 20 ~ 2048
                cloud_ssd - 20 ~ 2048
                The value should be equal to or greater than the size of the specific SnapshotId.

        :type disk_tags: list
        :param disk_tags: A list of hash/dictionaries of instance
            tags, '[{tag_key:"value", tag_value:"value"}]', tag_key
            must be not null when tag_value isn't null        

        :type snapshot_id: integer
        :param snapshot_id: Snapshots are used to create the data disk
            After this parameter is specified, Size is ignored.

        :rtype: dict
        :return: Returns a dictionary of disk information
        """
        return self.describe_disk_attribute(disk_id=self.get_object_new(self.build_request_params(self.format_request_kwargs(**kwargs)), ResultSet).disk_id)

    def describe_disk_attribute(self, disk_id):
        disks = self.describe_disks(disk_ids=[disk_id])
        if disks:
            return disks[0]
        return None

    def attach_disk(self, **kwargs):
        """
        Method to attach a disk to instance

        :type instance_id: string
        :param instance_id: The instance's ID

        :type disk_id: string
        :param disk_id: The disk ID in the cloud

        :type delete_with_instance: string
        :param delete_with_instance: value depicting should disk be deleted with instance.

        :return: A list of the total number of security groups, region ID of the security group,
                 the ID of the VPC to which the security group belongs
        """
        self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))
        return self.wait_for_disk_status(disk_id=kwargs["disk_id"], status="in_use", delay=5, timeout=120)

    def detach_disk(self, **kwargs):
        """
        Method to detach a disk to instance

        :type disk_id: dict
        :param disk_id: ID of Disk for attaching detaching disk

        :return: Return status of Operation
        """
        self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))
        return self.wait_for_disk_status(disk_id=kwargs["disk_id"], status="available", delay=5, timeout=120)

    def delete_disk(self, **kwargs):
        """
        Method to delete a disk

        :type disk_id: dict
        :param disk_id: ID of Disk for attaching detaching disk

        :return: Return status of Operation
        """
        kwargs = self.format_request_kwargs(**kwargs)

        delay = 3
        timeout = DefaultTimeOut

        while True:
            try:
                self.get_status_new(self.build_request_params(kwargs))
                volume = self.describe_disk_attribute(disk_id=kwargs["disk_id"])
                if volume is None or not volume.id:
                    return True
            except ServerException as e:
                if str(e.error_code) == "InvalidDiskId.NotFound":
                    return True
                elif str(e.error_code) == "IncorrectInstanceStatus.Initializing":
                    pass
            time.sleep(delay)
            timeout -= delay
            if timeout <= 0:
                raise Exception("Timeout: Waiting for deleting volume {0}, time-consuming {1} seconds. "
                                "Error: {2}".format(kwargs["disk_id"], DefaultTimeOut, e))

    def modify_disk_attribute(self, **kwargs):
        """
        Method to delete a disk

        :type disk_id: dict
        :param disk_id: ID of Disk for attaching detaching disk

        :return: Return status of Operation
        """
        return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))

    def create_image(self, snapshot_id=None, image_name=None, image_version=None, description=None,
                     instance_id=None, disk_mapping=None, client_token=None, wait=None,
                     wait_timeout=None):
        """
        Create a user-defined image with snapshots of system disks.
        The created image can be used to create a new ECS instance.

        :type snapshot_id: string
        :param snapshot_id: A user-defined image is created from the specified snapshot.

        :type image_name: string
        :param image_name: image name which is to be created

        :type image_version: string
        :param image_version: version of image

        :type description: string
        :param description: description of the image

        :type instance_id: string
        :param instance_id: the specified instance_id

        :type disk_mapping: list
        :param disk_mapping: An optional list of device hashes/dictionaries with custom configurations
      
        :type client_token: string
        :param client_token: An optional list of device hashes/dictionaries with custom configurations

        :type wait: string
        :param wait: An optional bool value indicating wait for instance to be running before running

        :type wait_timeout: int
        :param wait_timeout: An optional int value indicating how long to wait, default 300

        :return: Image id
        """
        params = {}
        results = []
        changed = False
        image_id = ''
        request_id = ''

        # the snapshot id for creating image
        if snapshot_id:
            # Verifying progress of snapshot_id, snapshot_id should be 100% completed
            snapshot_results, snapshot_progress, snapshot_changed = self.get_snapshot_image(snapshot_id)

            if snapshot_results:
                if 'error code' in str(snapshot_results).lower():
                    results = snapshot_results
                    return changed, image_id, results, request_id

            if not snapshot_changed:
                results.append({"Error Code": "Snapshot.NotReady", "Error Message": "snapshot is not ready"})
                return changed, image_id, results, request_id

        if snapshot_id:
            self.build_list_params(params, snapshot_id, 'SnapshotId')

        # set the image name
        if image_name:
            self.build_list_params(params, image_name, 'ImageName')

        # set the image version
        if image_version:
            self.build_list_params(params, image_version, 'ImageVersion')

        # set the description
        if description:
            self.build_list_params(params, description, 'Description')

        # set the client token
        if client_token:
            self.build_list_params(params, client_token, 'ClientToken')

        # specify the instance id
        if instance_id:
            self.build_list_params(params, instance_id, 'InstanceId')

        # specify the disk device mapping, An optional list of device hashes/dictionaries with custom configurations
        if disk_mapping:
            mapping_no = 1
            for mapping in disk_mapping:
                if mapping:
                    if 'disk_size' in mapping:
                        self.build_list_params(params, mapping[
                            'disk_size'], 'DiskDeviceMapping.' + str(mapping_no) + '.Size')
                    if 'snapshot_id' in mapping:
                        self.build_list_params(params, mapping[
                            'snapshot_id'], 'DiskDeviceMapping.' + str(mapping_no) + '.SnapshotId')
                        snapshot_map_results, snapshot_map_progress, snapshot_map_changed \
                            = self.get_snapshot_image(mapping['snapshot_id'])
                        if snapshot_map_results:
                            if 'error code' in str(snapshot_map_results).lower():
                                results = snapshot_map_results
                                return changed, image_id, results, request_id

                        if not snapshot_map_changed:
                            results.append(
                                {"Error Code": "Snapshot.NotReady", "Error Message": "snapshot is not ready"})
                            return changed, image_id, results, request_id

                    mapping_no += 1

        try:
            response = self.get_object('CreateImage', params, ResultSet)

            if response:
                image_id = response.image_id
                request_id = response.request_id

            if wait:
                if not wait_timeout:
                    wait_timeout = 300
                time.sleep(wait_timeout)

            results.append("Image creation successful")
            changed = True
         
        except ServerException as e:
            results.append({"Error Code": e.error_code, "Error Message": e.message,
                            "RequestId": e.request_id, "Http Status": e.http_status})
        except Exception as e:
            results.append({"Error:": e})

        return changed, image_id, results, request_id

    def delete_image(self, image_id):
        """
        Delete image , delete image inside particular region.
        :type image_id: dict
        :param image_id: ID of an Image        
        :rtype: Return status of Operation
        """
        params = {}
        results = []
        changed = False

        self.build_list_params(params, image_id, 'ImageId')

        try:
            response = self.get_object('DescribeImages', params, ResultSet)
            if response:
                json_obj = response
                total_instance = response.total_count
                if total_instance > 0:
                    for items in response.images['image']:
                        if image_id == items['image_id']:
                            response = self.get_status('DeleteImage', params)
                            results.append(response)
                            changed = True
                else:
                    results.append({"Error Code": "Image does not exist", "Error Message": "Image does not exist"})
        except ServerException as e:
            results.append({"Error Code": e.error_code, "Error Message": e.message,
                            "RequestId": e.request_id, "Http Status": e.http_status})
            changed = False
        except Exception as e:
            results.append({"Error:": e})
            changed = False

        return changed

    def get_all_images(self, image_id=None, image_name=None, 
                           snapshot_id=None, filters=None):
            """
            Get all Volumes associated with the current credentials.

            :type image_id: dict
            :param image_id: Optional  image id.  If this is present,
                             only the image associated with these
                             image id will be returned.
        
            :type image_name: dict
            :param image_name: Optional image name.  If this is present,
                               only the image associated with these 
                               image name will be returned.

            :type snapshot_id: list
            :param snapshot_id: Optional snapshot id.  If this list is
                                present, only the image associated with
                                these snapshot id will be returned.

            :type filters: dict
            :param filters: Optional filters that can be used to limit
                            the results returned.  Filters are provided
                            in the form of a dictionary consisting of
                            filter names as the key and filter values
                            as the value.  The set of allowable filter
                            names/values is dependent on the request
                            being performed.  Check the ECS API guide
                            for details.

            :rtype: list of Volume
            :return: The requested Volume objects
            """
            params = {}
            result = []
            changed = False
            if image_id:
                self.build_list_params(params, image_id, 'ImageId')
            if image_name:
                self.build_list_params(params, image_name, 'ImageName')
            if snapshot_id:
                self.build_list_params(params, snapshot_id, 'SnapshotId')
            if filters:
                self.build_filter_params(params, filters)
            result = self.get_list('DescribeImages', params, ['Images', Image])
            if result:
                changed = True;
            else:
                while changed == False:
                    time.sleep(20);
                    result = self.get_list('DescribeImages', params, ['Images', Image])
                    if result:
                        changed = True;
                        break                
                    
            return result;

    def get_snapshot_image(self, snapshot_id):
        params = {}
        results = []
        progress = ''
        changed = False
        counter = 0
        try:
            while changed == False:
                self.build_list_params(params, [snapshot_id], 'SnapshotIds')

                if counter > 20:
                    break
                obtained_results = self.get_object('DescribeSnapshots', params, ResultSet)
                counter += 1
                if obtained_results and len(obtained_results.snapshots[u'snapshot']) > 0:
                    status = str(obtained_results.snapshots[u'snapshot'][0][u'status'])
                    progress = str(obtained_results.snapshots[u'snapshot'][0][u'progress'])

                    if not '100%' in progress:
                        time.sleep(60)
                    else:
                        changed = True
                        progress = '100'
                        break
                else:
                    results.append({"Error Code": "Invalid.SnapshotId", "Error Message": "The snapshot id not found"})
                    break
        except ServerException as e:
            results.append({"Error Code": e.error_code, "Error Message": e.message,
                            "RequestId": e.request_id, "Http Status": e.http_status})
        except Exception as e:
            results.append({"Error:": e})

        return results, progress, changed

    def get_instance_details(self, instance_id):
        """
        Get details of an Instance
        :param instance_id: Id of an Instance
        :return: Return info about instance
        """
        params = {}

        self.build_list_params(params, instance_id, 'InstanceId')

        return self.get_object('DescribeInstanceAttribute', params, Instance)

    def wait_for_instance_status(self, instance_id, status, delay=DefaultWaitForInterval, timeout=DefaultTimeOut):
        """
        To verify instance status has become expected
        """
        tm = timeout
        try:
            while True:
                instance = self.get_instance_details(instance_id)
                if instance and str(instance.status).lower() in [status, str(status).lower()]:
                    return True

                tm -= delay

                if tm <= 0:
                    raise Exception("Timeout Error: Waiting for Disk status is %s, time-consuming %d seconds." % (status, timeout))

                time.sleep(delay)
            return False
        except Exception as e:
            raise e

    def wait_for_disk_status(self, disk_id, status, delay=DefaultWaitForInterval, timeout=DefaultTimeOut):
        """
        To verify disk status has become expected after attaching or detaching disk
        """
        tm = timeout
        try:
            while True:
                volume = self.describe_disk_attribute(disk_id)
                if volume and str(volume.status).lower() in [status, str(status).lower()]:
                    return True

                tm -= delay

                if tm <= 0:
                    raise Exception("Timeout Error: Waiting for Disk status is %s, time-consuming %d seconds." % (status, timeout))

                time.sleep(delay)
            return False
        except Exception as e:
            raise e

    def delete_instance_retry(self, action, params, instance_id, delay=DefaultWaitForInterval, timeout=DefaultTimeOut):
        while True:
            try:
                self.get_status(action, params)
                instance = self.get_instance_details(instance_id)
                if instance is None or not instance.id:
                    return True
            except ServerException as e:
                if str(e.error_code) == "InvalidInstanceId.NotFound":
                    return True

            time.sleep(delay)
            timeout -= delay
            if timeout <= 0:
                raise Exception("Timeout Error: Waiting for deleting instance {0}, time-consuming {1} seconds. "
                                "Error: {2}".format(instance_id, DefaultTimeOut, e))

    def verify_join_remove_securitygrp(self, instance_id, group_id, mode, delay=DefaultWaitForInterval, timeout=DefaultTimeOut):
        """
        To verify join & remove operations got performed in security group
        """
        done = False
        count = 0
        id_of_instance = [instance_id]
        tm = timeout
        try:
            while not done:
                time.sleep(delay)
                instance_list = self.get_all_instances(id_of_instance, None, None)
                if len(instance_list) > 0:
                    if mode.lower() == 'join':
                        for inst in instance_list:
                            if len(inst.security_group_ids['security_group_id']) > 0:
                                for grp in inst.security_group_ids['security_group_id']:
                                    if str(grp) == group_id:
                                        done = True
                                        break

                    elif mode.lower() == 'remove':
                        for inst in instance_list:
                            if len(inst.security_group_ids['security_group_id']) > 0:
                                for grp in inst.security_group_ids['security_group_id']:
                                    if str(grp) == group_id:
                                        count = count + 1
                                if count == 0:
                                    done = True
                                    break
                tm -= delay
                if tm <= 0:
                    raise Exception("Timeout Error: Waiting for joining or removing security group, time-consuming %d seconds." % timeout)

        except Exception as ex:
            raise ex

        return done

    def get_all_regions(self):
        all_regions = self.get_list('DescribeRegions', None, ['Regions', RegionInfo])
        return all_regions

    def create_network_interface(self, params):
        params['Action'] = 'CreateNetworkInterface'
        result = self.get_object_new(self.build_request_params(params), ResultSet)
        if not self.wait_for_network_interface(result.network_interface_id, "Available"):
            raise Exception("Waitting Network Interface {0} Failed.".format("Available"))
        return self.get_network_interface(result.network_interface_id)

    def get_all_network_interfaces(self, filters=None):
        if not filters:
            filters = {}
        filters['Action'] = 'DescribeNetworkInterfaces'
        return self.get_list_new(self.build_request_params(filters), ['NetworkInterfaceSets', NetworkInterfaceSet])

    def get_network_interface(self, network_interface_id):
        result = self.get_all_network_interfaces({"network_interface_ids": [network_interface_id]})
        if len(result) == 1:
            return result[0]
        return None

    def attach_network_interface(self, network_interface_id, instance_id):
        params = {'NetworkInterfaceId': network_interface_id,
                  'InstanceId': instance_id,
                  'Action': 'AttachNetworkInterface'
                  }

        changed = self.get_status_new(self.build_request_params(params))
        if not self.wait_for_network_interface(network_interface_id, "InUse"):
            raise Exception("Waitting Network Interface {0} Failed.".format("InUse"))
        return changed

    def detach_network_interface(self, network_interface_id, instance_id):
        params = {'NetworkInterfaceId': network_interface_id,
                  'InstanceId': instance_id,
                  'Action': 'DetachNetworkInterface'
                  }

        changed = self.get_status_new(self.build_request_params(params))
        if not self.wait_for_network_interface(network_interface_id, "Available"):
            raise Exception("Waitting Network Interface {0} Failed.".format("Available"))
        return changed

    def modify_network_interface(self, params):
        params['Action'] = 'ModifyNetworkInterfaceAttribute'
        if self.get_status_new(self.build_request_params(params)):
            time.sleep(8)
            return True
        return False

    def delete_network_interface(self, network_interface_id):
        params = {'network_interface_id': network_interface_id,
                  'Action': 'DeleteNetworkInterface'
                  }
        return self.get_status_new(self.build_request_params(params))

    def wait_for_network_interface(self, id, status, delay=DefaultWaitForInterval, timeout=DefaultTimeOut):
        """
        To verify network interface status has become expected
        """
        tm = timeout
        while True:
            result = self.get_network_interface(id)
            if result and str(result.status).lower() in [status, str(status).lower()]:
                return True

            tm -= delay

            if tm < 0:
                raise Exception("Timeout Error: Waiting for network interface {0} {1}, time-consuming {2} seconds.".format(id, status, timeout))

            time.sleep(delay)
        return False

    def add_tags(self, **kwargs):
        return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))

    def remove_tags(self, **kwargs):
        return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))


class image(object):
    """description of class"""

"""
Represents an ECS Image
"""
from footmark.ecs.ecsobject import TaggedECSObject


class Image(TaggedECSObject):
    def __init__(self, connection=None):
        super(Image, self).__init__(connection)
        self.tags = {}

    def __repr__(self):
        return 'Image:%s' % self.id

    def __getattr__(self, name):
        if name == 'id':
            return self.image_id
        if name == 'name':
            return self.image_name
        raise AttributeError

    def __setattr__(self, name, value):
        if name == 'id':
            self.image_id = value
        if name == 'name':
            self.image_name = value
        if name == 'tags' and value:
            v = {}
            for tag in value['tag']:
                v[tag.get('TagKey')] = tag.get('TagValue', None)
            value = v
        super(TaggedECSObject, self).__setattr__(name, value)

    def delete(self):
        """
        Terminate the image
        """
        return self.connection.delete_image(self.id)






"""
Represents an ECS Instance
"""
from footmark.ecs.ecsobject import TaggedECSObject


class Instance(TaggedECSObject):
    """
    Represents an instance.
    """

    def __init__(self, connection=None):
        super(Instance, self).__init__(connection)
        self.tags = {}

    def __repr__(self):
        return 'Instance:%s' % self.id

    def __getattr__(self, name):
        if name == 'id':
            return self.instance_id
        if name == 'name':
            return self.instance_name
        if name == 'state':
            return self.status
        if name in ('inner_ip', 'inner_ip_address'):
            return self.inner_ip_address['ip_address'][0]
        if name in ('public_ip', 'assign_public_ip', 'public_ip_address'):
                return self.public_ip_address['ip_address'][0]
        if name in ('private_ip', 'private_ip_address', 'vpc_private_ip', 'vpc_private_ip_address'):
            return self.vpc_attributes['private_ip_address']['ip_address'][0]
        if name in ('vpc_vswitch_id', 'vswitch_id', 'vpc_subnet_id', 'subnet_id'):
            return self.vpc_attributes['vswitch_id']
        if name == 'vpc_id':
            return self.vpc_attributes['vpc_id']
        if name in ('eip', 'elastic_ip_address'):
            return self.eip_address
        if name in ('group_id', 'security_group_id'):
            return self.security_group_id
        if name in ('group_name', 'security_group_name') and self.security_groups:
            return self.security_groups[0].security_group_name
        if name == 'groups':
            return self.security_groups
        if name in ('key_name', 'keypair', 'key_pair'):
            return getattr(self, 'key_pair_name', '')
        raise AttributeError("Object {0} does not have attribute {1}".format(self.__repr__(), name))

    def __setattr__(self, name, value):
        if name == 'id':
            self.instance_id = value
        if name == 'name':
            self.instance_name = value
        if name == 'status':
            value = value.lower()
        if name == 'state':
            self.status = value
        if name in ('public_ip_address', 'inner_ip_address', 'private_ip_address'):
            if isinstance(value, dict):
                if value['ip_address']:
                    value = value['ip_address'][0]
                else:
                    value = None
        if name  == 'eip_address' and isinstance(value, dict) and value['ip_address']:
            value = value['ip_address']
        if name == 'inner_ip':
            self.inner_ip_address = value
        if name in ('public_ip', 'assign_public_ip'):
            self.public_ip_address = value
        if name in ('private_ip', 'vpc_private_ip', 'vpc_private_ip_address'):
            self.vpc_attributes['private_ip_address'] = value
            self.private_ip_address = value
        if name in ('vpc_vswitch_id', 'vswitch_id', 'vpc_subnet_id', 'subnet_id'):
            self.vpc_attributes['vswitch_id'] = value
        if name == 'vpc_id':
            self.vpc_attributes['vpc_id'] = value
        if name in ('eip', 'elastic_ip_address'):
            self.eip_address = value
        if name in ('group_id', 'security_group_id'):
            if isinstance(value, list) and value:
                value = value[0]
        if name in ('group_name', 'security_group_name') and self.security_groups:
            self.security_groups[0].security_group_name = value
        if name == 'groups':
            self.security_groups = value
        if name in ('key_name', 'keypair', 'key_pair'):
            self.key_pair_name = value
        if name == 'tags' and value:
            v = {}
            for tag in value['tag']:
                if tag.get('tag_key'):
                    v[tag.get('tag_key')] = tag.get('tag_value', None)
            value = v
        super(TaggedECSObject, self).__setattr__(name, value)

    def _update(self, updated):
        self.__dict__.update(updated.__dict__)

    def update(self, validate=False):
        """
        Update the instance's state information by making a call to fetch
        the current instance attributes from the service.

        :type validate: bool
        :param validate: By default, if ECS returns no data about the
                         instance the update method returns quietly.  If
                         the validate param is True, however, it will
                         raise a ValueError exception if no data is
                         returned from ECS.
        """
        rs = self.connection.describe_instances(instance_ids=[self.id])
        if len(rs) > 0:
            for r in rs:
                if r.id == self.id:
                    self._update(r)
        elif validate:
            raise ValueError('%s is not a valid Instance ID' % self.id)
        return self.state

    def start(self):
        """
        Start the instance.
        """
        return self.connection.start_instances(instance_ids=[self.id])

    def stop(self, force=False):
        """
        Stop the instance

        :type force: bool
        :param force: Forces the instance to stop

        :rtype: list
        :return: A list of the instances stopped
        """
        return self.connection.stop_instances(instance_ids=[self.id], force_stop=force)

    def reboot(self, force=False):
        """
        Restart the instance.

        :type force: bool
        :param force: Forces the instance to stop
        """
        return self.connection.reboot_instances(instance_ids=[self.id], force_stop=force)

    def modify(self, name=None, description=None, host_name=None, password=None, user_data=None):
        """
        Modify the instance.

        :type name: str
        :param name: Instance Name
        :type description: str
        :param description: Instance Description
        :type host_name: str
        :param host_name: Instance Host Name
        :type password: str
        :param password: Instance Password
        """
        params = {}
        if name and self.name != name:
            params['instance_name'] = name
        if description and self.description != description:
            params['description'] = description
        if host_name and self.host_name != host_name:
            params['host_name'] = host_name
        if user_data and self.user_data != user_data:
            params['user_data'] = user_data
        if password:
            params['password'] = password
        if params:
            params['instance_id'] = self.instance_id
            return self.connection.modify_instance_attribute(**params)
        return False

    def terminate(self, force=False):
        """
        Terminate the instance

        :type force: bool
        :param force: Forces the instance to terminate
        """
        return self.connection.delete_instances(instance_ids=[self.id], force=force)

    def join_security_group(self, security_group_id):
        """
        Join one security group

        :type security_group_id: str
        :param security_group_id: The Security Group ID.
        """
        return self.connection.join_security_group(self.id, security_group_id)

    def leave_security_group(self, security_group_id):
        """
        Leave one security group

        :type security_group_id: str
        :param security_group_id: The Security Group ID.
        """
        return self.connection.leave_security_group(self.id, security_group_id)

    def attach_key_pair(self, key_pair_name):
        """
        Attach one key pair

        :type key_pair_name: str
        :param key_pair_name: The Key Pair Name.
        """
        return self.connection.attach_key_pair([self.id], key_pair_name)

    def detach_key_pair(self):
        """
        detach one key pair
        """
        return self.connection.detach_key_pair([self.id], self.key_pair_name)

    def add_tags(self, tags):
        """
        Add tags
        """
        remain = {}
        if tags:
            for key, value in tags.items():
                if key in self.tags.keys() and value == self.tags[key]:
                    continue
                remain[key] = value
        if remain:
            return self.connection.add_tags(resource_id=self.id, resource_type="instance", tags=remain)
        return False

    def remove_tags(self, tags):
        """
        remove tags
        """
        remain = {}
        if tags:
            for key, value in tags.items():
                if key not in self.tags.keys():
                    continue
                remain[key] = value
        if remain:
            return self.connection.remove_tags(resource_id=self.id, resource_type="instance", tags=tags)
        return False

    def allocate_public_ip(self):
        if self.public_ip_address:
            return False
        return self.connection.allocate_public_ip_address(instance_id=self.id)

    def read(self):
        instance = {"gpu": {"amount": 0, "spec": ""}}
        for name, value in self.__dict__.items():
            if name in ["connection", "region_id", "region", "security_group_ids", "dedicated_host_attribute", "device_available",
                        "instance_type_family", "operation_locks", "recyclable", "sale_cycle", "serial_number", "stopped_mode",
                        "vlan_id", "spot_price_limit", "spot_strategy", "cluster_id", "instance_network_type", "start_time"]:
                continue

            if name == "instance_id":
                instance['id'] = value

            if name == "block_device_mapping":
                volumes = []
                for disk in value:
                    volumes.append({
                        "device_name": disk.device,
                        "attach_time": disk.attached_time,
                        "delete_on_termination": disk.delete_with_instance,
                        "status": disk.status,
                        "volume_id": disk.disk_id
                    })
                value = volumes

            if name == "security_groups":
                groups = []
                for sg in value:
                    groups.append({
                        "group_id": sg.group_id,
                        "group_name": sg.group_name
                    })
                value = groups

            if name == "vpc_attributes":
                instance["vpc_id"] = value["vpc_id"]
                instance["vswitch_id"] = value["vswitch_id"]
                if value["private_ip_address"]["ip_address"]:
                    instance["private_ip_address"] = value["private_ip_address"]["ip_address"][0]
                continue

            if name == "network_interfaces":
                value = value["network_interface"]

            # if name == "public_ip_address":
            #     print "******* {0}".format(value)
            #     if value and value["ip_address"]:
            #         value = value["ip_address"][0]

            if name == "inner_ip_address":
                if value and value["ip_address"]:
                    value = value["ip_address"][0]

            if name == "eip_address":
                name = "eip"
                if value["ip_address"]:
                    value["ip_address"] = value["ip_address"][0]

            if name == "gpuamount":
                instance["gpu"]["amount"] = value
                continue

            if name == "gpuspec":
                instance["gpu"]["specification"] = value
                continue

            if name == "zone_id":
                name = "availability_zone"

            if name == "statue":
                name = "state"

            instance[name] = value
        return instance



from footmark.regioninfo import RegionInfo


class OSSRegionInfo(RegionInfo):
    """
    Represents an OSS Region
    """

    def __init__(self, connection=None, name=None, id=None,
                 connection_cls=None):
        from footmark.oss.connection import OSSConnection
        super(OSSRegionInfo, self).__init__(connection, name, id,
                                            OSSConnection)



"""
Represents an VPC Security Group
"""
from footmark.vpc.vpcobject import TaggedVPCObject


class Vpc(TaggedVPCObject):
    def __init__(self, connection=None, owner_id=None,
                 name=None, description=None, id=None):
        super(Vpc, self).__init__(connection)
        self.tags = {}

    def __repr__(self):
        return 'Vpc:%s' % self.id

    def __getattr__(self, name):
        if name == 'id':
            return self.vpc_id
        if name == 'name':
            return self.vpc_name
        if name == 'router_id':
            return self.vrouter_id
        raise AttributeError

    def __setattr__(self, name, value):
        if name == 'id':
            self.vpc_id = value
        if name == 'name':
            self.vpc_name = value
        if name == 'router_id':
            self.vrouter_id = value
        if name == 'tags' and value:
            v = {}
            for tag in value['tag']:
                v[tag.get('TagKey')] = tag.get('TagValue', None)
            value = v
        super(TaggedVPCObject, self).__setattr__(name, value)

    def modify(self, vpc_name=None, description=None):
        """
        Update vpc's attribute
        """
        params = {}
        if vpc_name and self.name != vpc_name:
            params['vpc_name'] = vpc_name
        if description and self.description != description:
            params['description'] = description
        if params:
            params['vpc_id'] = self.vpc_id
            return self.connection.modify_vpc_attribute(**params)
        return False

    def get(self):
        return self.connection.describe_vpc_attribute(vpc_id=self.id)

    def read(self):
        vpc = {}
        for name, value in self.__dict__.items():
            if name in ["connection", "region_id", "region"]:
                continue

            if name == 'vpc_id':
                vpc['id'] = value

            if name == 'status':
                name = 'state'
                value = str(value).lower()

            if name == 'user_cidrs':
                value = value['user_cidr']

            if name == 'vswitch_ids':
                value = value['vswitch_id']

            vpc[name] = value
        return vpc

    def delete(self):
        """
        Terminate the vpc
        """
        return self.connection.delete_vpc(vpc_id=self.id)



"""
Represents an ECS Instance Type
"""
from footmark.ecs.ecsobject import *


class InstanceType(TaggedECSObject):
    """
    Represents an instance type.
    """

    def __init__(self, connection=None):
        super(InstanceType, self).__init__(connection)
        self.tags = {}

    def __repr__(self):
        return 'InstanceType:%s' % self.id

    def __getattr__(self, name):
        if name == 'id':
            return self.instance_type_id
        if name == 'family':
            return self.instance_type_family
        raise AttributeError

    def __setattr__(self, name, value):
        if name == 'id':
            self.instance_type_id = value
        if name == 'family':
            self.instance_type_family = value
        super(TaggedECSObject, self).__setattr__(name, value)


class InstanceTypeFamily(TaggedECSObject):
    """
       Represents an instance type family.
    """

    def __init__(self, connection=None):
        super(InstanceTypeFamily, self).__init__(connection)
        self.tags = {}

    def __repr__(self):
        return 'InstanceTypeFamily:%s' % self.id

    def __getattr__(self, name):
        if name == 'id':
            return self.instance_type_family_id
        raise AttributeError

    def __setattr__(self, name, value):
        if name == 'id':
            self.instance_type_family_id = value
        super(TaggedECSObject, self).__setattr__(name, value)




"""
Represents an ECS Instance
"""
from footmark.ess.essobject import TaggedESSObject


class ScalingGroup(TaggedESSObject):
    """
    Represents an scaling configuration.
    """

    def __init__(self, connection=None):
        super(ScalingGroup, self).__init__(connection)
        self.tags = {}

    def __repr__(self):
        return 'Scaling Group:%s' % self.id

    def __getattr__(self, name):
        if name == 'id':
            return self.scaling_group_id
        if name == 'name':
            return self.scaling_group_name
        if name in ('state', 'status'):
            return self.lifecycle_state
        if name in ('configuration_id', 'scaling_configuration_id'):
            return getattr(self, 'active_scaling_configuration_id', None)
        if name == 'cooldown':
            return self.default_cooldown
        if name in ('vswitch_id', 'subnet_id'):
            return getattr(self, 'v_switch_id', None)
        if name in ('vswitch_ids', 'subnet_ids'):
            return getattr(self, 'v_switch_ids', None)
        raise AttributeError("Object {0} does not have attribute {1}".format(self.__repr__(), name))

    def __setattr__(self, name, value):
        if name == 'id':
            self.scaling_group_id = value
        if name == 'name':
            self.scaling_group_name = value
        if name == 'lifecycle_state':
            value = value.lower()
        if name in ('state', 'status'):
            self.lifecycle_state = value
        if name in ('configuration_id', 'scaling_configuration_id'):
            self.active_scaling_configuration_id = value
        if name == 'cooldown':
            self.default_cooldown = value
        if name in ('vswitch_id', 'subnet_id'):
            self.v_switch_id = value
        if name in ('vswitch_ids', 'subnet_ids'):
            self.v_switch_ids = value
        if name == 'tags' and value:
            v = {}
            for tag in value['tag']:
                if tag.get('tag_key'):
                    v[tag.get('tag_key')] = tag.get('tag_value', None)
            value = v
        super(TaggedESSObject, self).__setattr__(name, value)

    def _update(self, updated):
        self.__dict__.update(updated.__dict__)

    def update(self, validate=False):
        """
        Update the instance's state information by making a call to fetch
        the current instance attributes from the service.scaling_group_ids=None, scaling_group_names=None

        :type validate: bool
        :param validate: By default, if ECS returns no data about the
                         instance the update method returns quietly.  If
                         the validate param is True, however, it will
                         raise a ValueError exception if no data is
                         returned from ECS.
        """
        rs = self.connection.describe_groups(scaling_group_ids=[self.id])
        if len(rs) > 0:
            for r in rs:
                if r.id == self.id:
                    self._update(r)
        elif validate:
            raise ValueError('%s is not a valid Scaling Configuration ID' % self.id)
        return self.state

    def enable(self, scaling_configuration_id=None, instance_ids=None):
        """
        Enable the scaling group.
        """
        return self.connection.enable_group(self.id, scaling_configuration_id=scaling_configuration_id, instance_ids=instance_ids)

    def disable(self):
        """
        Disable the scaling group

        :type force: bool
        :param force: Forces the instance to stop

        :rtype: list
        :return: A list of the instances stopped
        """
        return self.connection.disable_group(self.id)

    def modify(self, max_size=None, min_size=None, name=None, default_cooldown=None, removal_policies=None, scaling_configuration_id=None):
        """
        Stop the instance

        :type force: bool
        :param force: Forces the instance to stop

        :rtype: list
        :return: A list of the instances stopped
        """
        if min_size:
            if not max_size:
                max_size = self.max_size
            if min_size > max_size:
                min_size = max_size

        return self.connection.modify_group(self.id, max_size=max_size, min_size=min_size, name=name, default_cooldown=default_cooldown,
                                            removal_policies=removal_policies, scaling_configuration_id=scaling_configuration_id)

    def terminate(self, force=False):
        """
        Terminate the instance

        :type force: bool
        :param force: Forces the instance to terminate
        """
        return self.connection.terminate_group(self.id, force=force)



from footmark.rds.rdsobject import TaggedRDSObject

class DbInstance(TaggedRDSObject):
    def __init__(self, connection=None):
        super(DbInstance, self).__init__(connection)

    def __repr__(self):
        return 'DbInstance:%s' % self.id

    def __getattr__(self, name):
        if name == 'id':
             return self.id
        if name == 'status':
             return self.dbinstance_status
        if name == 'type':
             return self.dbinstance_type

    def __setattr__(self, name, value):
        if name == 'id':
             self.dbinstance_id=value
        if name == 'status':
             self.dbinstance_status=value
        if name == 'type':
             self.dbinstance_type=value
        super(TaggedRDSObject, self).__setattr__(name, value)


class Account(TaggedRDSObject):
    def __init__(self, connection=None, owner_id=None,
                 name=None, description=None, id=None):
        super(Account, self).__init__(connection)
        self.tags = {}

    def __repr__(self):
        return 'Account:%s' % self.name

    def __getattr__(self, name):
        if name == 'name':
             return self.account_name
        if name == 'status':
             return self.account_status
        if name == 'description':
             return self.account_description
        if name == 'privileges':
             return self.database_privileges

    def __setattr__(self, name, value):
        if name == 'name':
             self.account_name=value
        if name == 'status':
             self.account_status=value
        if name == 'description':
             self.account_description=value
        if name == 'privileges':
             self.database_privileges=value
        super(TaggedRDSObject, self).__setattr__(name, value)
    
    def reset(self, dbinstance_id, account_password):
        '''
        reset
        '''
        return self.connection.reset_account_password(dbinstance_id, self.account_name, account_password)
    
    def grant_privilege(self, dbinstance_id, db_name, account_privilege):
        '''
        grant privilege
        '''
        return self.connection.grant_account_privilege(dbinstance_id, self.account_name, db_name, account_privilege)
    
    def revoke_privilege(self, dbinstance_id, db_name):
        '''
        revoke privilege
        '''
        return self.connection.revoke_account_privilege(dbinstance_id, self.account_name, db_name)
    
    def modify_description(self, dbinstance_id, description):
        '''
        modify description
        '''
        return self.connection.modify_account_description(dbinstance_id, self.account_name, description)
    
    def delete(self, dbinstance_id):
        '''
        delete account 
        '''
        return self.connection.delete_account(dbinstance_id, self.account_name)



# encoding: utf-8
import warnings
import time

from footmark.connection import ACSQueryConnection
from footmark.vpc.regioninfo import RegionInfo
from footmark.exception import VPCResponseError
from footmark.resultset import ResultSet
from footmark.vpc.vpc import Vpc
from footmark.vpc.eip import Eip
from footmark.vpc.vswitch import VSwitch
from footmark.vpc.router import RouteEntry, RouteTable
from footmark.vpc.config import *
from aliyunsdkcore.acs_exception.exceptions import ServerException
# from aliyunsdkvpc.request.v20160428.DescribeVpcAttributeRequest import


class VPCConnection(ACSQueryConnection):
    SDKVersion = '2016-04-28'
    DefaultRegionId = 'cn-hangzhou'
    DefaultRegionName = u'杭州'.encode("UTF-8")
    ResponseError = VPCResponseError

    def __init__(self, acs_access_key_id=None, acs_secret_access_key=None,
                 region=None, sdk_version=None, security_token=None, user_agent=None):
        """
        Init method to create a new connection to ECS.
        """
        if not region:
            region = RegionInfo(self, self.DefaultRegionName,
                                self.DefaultRegionId)
        self.region = region
        if sdk_version:
            self.SDKVersion = sdk_version

        self.VPCSDK = 'aliyunsdkvpc.request.v' + self.SDKVersion.replace('-', '')

        super(VPCConnection, self).__init__(acs_access_key_id=acs_access_key_id,
                                            acs_secret_access_key=acs_secret_access_key,
                                            region=self.region, product=self.VPCSDK,
                                            security_token=security_token, user_agent=user_agent)

    def build_filter_params(self, params, filters):
        if not isinstance(filters, dict):
            return

        flag = 1
        for key, value in filters.items():
            acs_key = key
            if acs_key.startswith('tag:'):
                while ('set_Tag%dKey' % flag) in params:
                    flag += 1
                if flag < 6:
                    params['set_Tag%dKey' % flag] = acs_key[4:]
                    params['set_Tag%dValue' % flag] = filters[acs_key]
                flag += 1
                continue
            if key == 'group_id':
                if not value.startswith('sg-') or len(value) != 12:
                    warnings.warn("The group-id filter now requires a security group "
                                  "identifier (sg-*) instead of a security group ID. "
                                  "The group-id " + value + "may be invalid.",
                                  UserWarning)
                params['set_SecurityGroupId'] = value
                continue
            if not isinstance(value, dict):
                acs_key = ''.join(s.capitalize() for s in acs_key.split('_'))
                params['set_' + acs_key] = value
                continue

            self.build_filters_params(params, value)

    def format_vpc_request_kwargs(self, **kwargs):
        for key, value in kwargs.items():

            # Format Vswitch to VSwitch
            if key == 'Action':
                if str(value).find("Vswitch"):
                    kwargs[key] = str(value).replace("Vswitch", "VSwitch")

            # Convert vpc user cidrs and vpc id
            if key in ['user_cidr', 'vpc_id']:
                if value:
                    if not isinstance(value, list):
                        value = [value]
                    kwargs[key] = str(",").join(value)

        return kwargs

    def create_vpc(self, **kwargs):
        vpc_id = self.get_object_new(self.build_request_params(self.format_vpc_request_kwargs(**self.format_request_kwargs(**kwargs))), ResultSet).vpc_id
        self.wait_for_vpc_status(vpc_id, 'Available', 4, 120)
        return self.describe_vpc_attribute(vpc_id=vpc_id)

    def describe_vpc_attribute(self, **kwargs):
        return self.get_object_new(self.build_request_params(self.format_request_kwargs(**kwargs)), Vpc)

    def describe_vpcs(self, **kwargs):
        return self.get_list_new(self.build_request_params(self.format_vpc_request_kwargs(**self.format_request_kwargs(**kwargs))), ['Vpcs', Vpc])

    def modify_vpc_attribute(self, **kwargs):
        if self.get_status_new(self.build_request_params(self.format_vpc_request_kwargs(**self.format_request_kwargs(**kwargs)))):
            return self.wait_for_vpc_status(kwargs['vpc_id'], 'Available', 4, 60)
        return False

    def delete_vpc(self, **kwargs):
        retry = 5
        while retry:
            try:
                return self.get_status_new(self.build_request_params(self.format_request_kwargs(**kwargs)))
            except ServerException as e:
                if str(e.error_code) == "Forbbiden" or str(e.error_code).find("Dependency"):
                    time.sleep(5)
                    retry -= 1
                    continue
                raise e
        return False

    def create_vswitch(self, **kwargs):
        vswitch_id = self.get_object_new(self.build_request_params(self.format_vpc_request_kwargs(**self.format_request_kwargs(**kwargs))), ResultSet).vswitch_id
        self.wait_for_vswitch_status(vswitch_id, 'Available', 4, 120)
        return self.describe_vswitch_attribute(vswitch_id=vswitch_id)

    def describe_vswitches(self, **kwargs):
        return self.get_list_new(self.build_request_params(self.format_vpc_request_kwargs(**self.format_request_kwargs(**kwargs))), ['VSwitches', VSwitch])

    def describe_vswitch_attributes(self, **kwargs):
        return self.get_object_new(self.build_request_params(self.format_vpc_request_kwargs(**self.format_request_kwargs(**kwargs))), VSwitch)

    # In order to unify, add an extra method. It is equal to describe_vswitch_attributes
    def describe_vswitch_attribute(self, **kwargs):
        kwargs = self.format_vpc_request_kwargs(**self.format_request_kwargs(**kwargs))
        kwargs['Action'] = kwargs['Action'] + 's'
        return self.get_object_new(self.build_request_params(kwargs), VSwitch)

    def modify_vswitch_attribute(self, **kwargs):
        return self.get_status_new(self.build_request_params(self.format_vpc_request_kwargs(**self.format_request_kwargs(**kwargs))))

    def delete_vswitch(self, **kwargs):
        return self.get_status_new(self.build_request_params(self.format_vpc_request_kwargs(**self.format_request_kwargs(**kwargs))))

    def create_route_entry(self, route_table_id, destination_cidrblock, nexthop_type=None, nexthop_id=None, nexthop_list=None):
        """
        Create RouteEntry for VPC
        :type route_table_id: str
        :param route_table_id: ID of route table in the VPC
        :type destination_cidrblock: str
        :param destination_cidrblock: The destination CIDR of route entry. It must be a legal CIDR or IP address, such as: 192.168.0.0/24 or 192.168.0.1
        :type nexthop_type: str
        :param nexthop_type: The type of next hop. Available value options: Instance, Tunnel, HaVip, RouterInterface. Default is Instance.
        :type next_hop_id: str
        :param next_hop_id: The ID of next hop.
        :type nexthop_list: str
        :param nexthop_list: The route item of next hop list. 
        :rtype 
        :return Return result of Creating RouteEntry.
        """
        params = {}

        self.build_list_params(params, route_table_id, 'RouteTableId')
        self.build_list_params(params, destination_cidrblock, 'DestinationCidrBlock')

        if nexthop_type:
            self.build_list_params(params, nexthop_type, 'NextHopType')

        if nexthop_id:
            self.build_list_params(params, nexthop_id, 'NextHopId')

        if nexthop_list:
            self.build_list_params(params, nexthop_list, 'NextHopList')

        if self.get_status('CreateRouteEntry', params):
            return self.wait_for_route_entry_status(route_table_id, destination_cidrblock, 'Available', 4, 60)

        return None

    def get_route_entry_attribute(self, route_table_id, destination_cidrblock, nexthop_id=None):
        """
        Querying route entry attribute
        :type route_table_id: str
        :param route_table_id: ID of route table in the VPC
        :type destination_cidrblock: str
        :param destination_cidrblock: The destination CIDR of route entry. It must be a legal CIDR or IP address, such as: 192.168.0.0/24 or 192.168.0.1
        :type nexthop_id: str
        :param nexthop_type: The ID of next hop.
        :rtype 
        :return: VRouters in json format
        """

        route_entries = self.get_all_route_entries(route_table_id=route_table_id)
        if route_entries:
            for entry in route_entries:
                if destination_cidrblock == str(entry.destination_cidrblock):
                    return entry
        return None

    def get_all_route_entries(self, router_id=None, router_type=None, route_table_id=None, pagenumber=1, pagesize=10):
        """
        Querying all route entries in the specified router or route_tables_id
        :type router_id: str
        :param router_id: The ID of router which is to be fetched.
        :type router_type str
        :param router_type: The type of router which is to be fetched.
        :type route_table_id: str
        :param route_table_id: ID of route table in one VPC
        :type pagenumber: integer
        :param pagenumber: Page number of the route table list. The start value is 1. The default value is 1
        :type pagesize: integer
        :param pagesize: Sets the number of lines per page for queries per page. The maximum value is 50.
        The default value is 10 
        :rtype list<>
        :return: List of route entry.
        """
        route_tables = self.get_all_route_tables(router_id=router_id, router_type=router_type, route_table_id=route_table_id,
                                                 pagenumber=pagenumber, pagesize=pagesize)
        route_entries = []
        if route_tables:
            for table in route_tables:
                if table.route_entrys:
                    for entry in table.route_entrys['route_entry']:
                        route_entry = RouteEntry(self)
                        for k, v in entry.items():
                            setattr(route_entry, k, v)
                        route_entries.append(route_entry)

        return route_entries

    def delete_route_entry(self, route_table_id, destination_cidrblock=None, nexthop_id=None, nexthop_list=None):
        """
        Deletes the specified RouteEntry for the vpc
        :type route_table_id: str
        :param route_table_id: ID of route table in the VPC
        :type destination_cidrblock: str
        :param destination_cidrblock: The destination CIDR of route entry. It must be a legal CIDR or IP address, such as: 192.168.0.0/24 or 192.168.0.1
        :type next_hop_id: str
        :param next_hop_id: The ID of next hop.
        :type nexthop_list: str
        :param nexthop_list: The route item of next hop list.
        :rtype bool
        :return Return result of deleting route entry.
        """
        params = {}

        self.build_list_params(params, route_table_id, 'RouteTableId')
        if destination_cidrblock:
            self.build_list_params(params, destination_cidrblock, 'DestinationCidrBlock')

        if nexthop_id:
            self.build_list_params(params, nexthop_id, 'NextHopId')

        if nexthop_list:
            self.build_list_params(params, nexthop_list, 'NextHopList')

        return self.get_status('DeleteRouteEntry', params)

    def get_route_table_attribute(self, route_table_id):
        """
        Querying route table attribute
        :type route_table_id: str
        :param route_table_id: ID of route table in the VPC
        :rtype dict
        :return: VRouters in json format
        """
        return self.get_all_route_tables(route_table_id=route_table_id)

    def get_all_route_tables(self, router_id=None, router_type=None, route_table_id=None, pagenumber=1, pagesize=10):
        """
        Querying vrouter
        :type router_id: str
        :param router_id: The ID of router which is to be fetched.
        :type router_type str
        :param router_type: The type of router which is to be fetched.
        :type route_table_id: str
        :param route_table_id: ID of route table in one VPC
        :type pagenumber: integer
        :param pagenumber: Page number of the route entry list. The start value is 1. The default value is 1
        :type pagesize: integer
        :param pagesize: Sets the number of lines per page for queries per page. The maximum value is 50.
        The default value is 10 
        :rtype list<>
        :return: List of route entry.
        """
        params = {}

        if router_id:
            self.build_list_params(params, router_id, 'RouterId')

        if router_type:
            self.build_list_params(params, router_type, 'RouterType')

        if route_table_id:
            self.build_list_params(params, route_table_id, 'RouteTableId')

        if pagenumber:
            self.build_list_params(params, pagenumber, 'PageNumber')

        if pagesize:
            self.build_list_params(params, pagesize, 'PageSize')

        return self.get_list('DescribeRouteTables', params, ['RouteTables', RouteTable])

    def get_instance_info(self):
        """
        method to get all Instances of particular region 
        :return: Return All Instances in the region
        """
        params = {}
        results = []

        try:
            v_ids = {}
            response = self.get_status('DescribeInstances', params)
            results.append(response)
            
        except Exception as ex:        
            error_code = ex.error_code
            error_msg = ex.message
            results.append({"Error Code": error_code, "Error Message": error_msg})

        return results

    def allocate_eip_address(self, bandwidth=5, internet_charge_type='PayByBandwidth', client_token=None):
        """
        method to query eip addresses in the region
        :type int
        :param bandwidth : bandwidth of the eip address. Default to 5
        :type internet_charge_type : str
        :param internet_charge_type : paybytraffic or paybybandwidth types
        :return: Return the allocationId , requestId and EIP address
        """
        params = {}
        self.build_list_params(params, str(bandwidth), 'Bandwidth')
        self.build_list_params(params, internet_charge_type, 'InternetChargeType')
        if client_token:
            self.build_list_params(params, client_token, 'ClientToken')
                  
        result = self.get_object('AllocateEipAddress', params, ResultSet)
        if result:
            return self.wait_for_eip_status(allocation_id=result.allocation_id, eip_address=result.eip_address,
                                            status='Available', interval=3, timeout=60)

        return None

    def get_all_eip_addresses(self, status=None, ip_address=None, allocation_id=None, associated_instance_type=None,
                              associated_instance_id=None, page_number=1, page_size=50):
        """
        Get EIP details for a region
        :param status: The EIP status includes Associating | Unassociating | InUse | Available
        :param ip_address: The EIP ip address
        :param allocation_id: ID of the allocated EIP
        :param associated_instance_type: The type of the associate device
        :param associated_instance_id: The ID of the associate device
        :param pagenumber: Page number. The start value is 1. The default value is 1
        :param pagesize: Sets the number of lines per page for queries per page. The maximum value is 50. Default to 50.
        :return:
        """
        params = {}

        if status:
            self.build_list_params(params, status, 'Status')
        if ip_address:
            self.build_list_params(params, ip_address, 'EipAddress')
        if allocation_id:
            self.build_list_params(params, allocation_id, 'AllocationId')
        if associated_instance_type:
            self.build_list_params(params, associated_instance_type, 'AssociatedInstanceType')
        if associated_instance_id:
            self.build_list_params(params, associated_instance_id, 'AssociatedInstanceId')

        self.build_list_params(params, page_number, 'PageNumber')
        self.build_list_params(params, page_size, 'PageSize')

        return self.get_list('DescribeEipAddresses', params, ['EipAddresses', Eip])

    def associate_eip(self, allocation_id, instance_id):
        """
        :type allocation_id:string
        :param allocation_id:The instance ID of the EIP
        :type instance_id:string
        :param instance_id:The ID of an ECS instance
        :param client_token: Used to ensure the idempotence of the request
        :return:Returns the status of operation
        """
        params = {}

        self.build_list_params(params, allocation_id, 'AllocationId')
        self.build_list_params(params, instance_id, 'InstanceId')
        if str(instance_id).startswith("lb-"):
            self.build_list_params(params, 'SlbInstance', 'InstanceType')
       
        self.get_status('AssociateEipAddress', params)
        return self.wait_for_eip_status(allocation_id=allocation_id, status="InUse", interval=2, timeout=60) is None

    def disassociate_eip(self, allocation_id, instance_id):
        """
        :type allocation_id:string
        :param allocation_id:The instance ID of the EIP
        :type instance_id:string
        :param instance_id:The ID of an ECS instance
        :return:Request Id
        """
        params = {}
        self.build_list_params(params, allocation_id, 'AllocationId')
        self.build_list_params(params, instance_id, 'InstanceId')
        self.get_status('UnassociateEipAddress', params)

        return self.wait_for_eip_status(allocation_id=allocation_id, status="Available", interval=2, timeout=60) is None

    def modify_eip(self, allocation_id, bandwidth):
        """
        :type allocation_id:string
        :param allocation_id:The instance ID of the EIP
        :type bandwidth:int
        :param bandwidth:Bandwidth of the EIP instance
        :return:Request Id
        """
        params = {}

        self.build_list_params(params, allocation_id, 'AllocationId')
        if int(bandwidth) > 0:
            self.build_list_params(params, int(bandwidth), 'Bandwidth')
        return self.get_status('ModifyEipAddressAttribute', params)

    def release_eip(self, allocation_id):
        """
        To release Elastic Ip
        :type allocation_id: string
        :param allocation_id: To release the allocation ID,allocation ID uniquely identifies the EIP
        :return: Return status of operation
        """
        params = {}

        self.build_list_params(params, allocation_id, 'AllocationId')
        try_times = 10
        while try_times > 0:
            self.get_status('ReleaseEipAddress', params)
            eips = self.get_all_eip_addresses(allocation_id=allocation_id)
            if not eips or len(eips) < 1:
                return True
            time.sleep(3)
            try_times -= 1

        raise Exception("Retry 10 times to release EIP failed."
                        "Please ensure EIP status is Available before releasing it.")

    def get_all_vrouters(self, vrouter_id=None, pagenumber=None, pagesize=None):
        """
        Querying vrouter
        :param vrouter_id: VRouter_Id to be fetched
        :type vrouter_id: str
        :type pagenumber: integer
        :param pagenumber: Page number of the instance status list. The start value is 1. The default value is 1
        :type pagesize: integer
        :param pagesize: Sets the number of lines per page for queries per page. The maximum value is 50.
        The default value is 10
        :return: VRouters in json format
        """
        params = {}
        results = []

        try:
            if vrouter_id is not None :
                self.build_list_params(params, vrouter_id, 'VRouterId')

            if pagenumber is not None :
                self.build_list_params(params, pagenumber, 'PageNumber')

            if pagesize is not None :
                self.build_list_params(params, pagesize, 'PageSize')

            results = self.get_status('DescribeVRouters', params)
        except Exception as ex:
            error_code = ex.error_code
            error_msg = ex.message
            results.append({"Error Code": error_code, "Error Message": error_msg})

        return False, results

    def wait_for_eip_status(self, allocation_id, status, eip_address=None, interval=DefaultWaitForInterval, timeout=DefaultTimeOut):
        """
        wait for bind ok
        :param eip_address:
        :param allocation_id:
        :param status:
        :return: 
        """
        try:
            tm = 0
            while tm < timeout:
                eips = self.get_all_eip_addresses(allocation_id=allocation_id, ip_address=eip_address)
                if not eips or len(eips) < 1:
                    return None
                if str.lower(status) == str.lower(eips[0].status):
                    return eips[0]
                tm += interval
                if tm >= timeout:
                    raise Exception("Timeout Error: Waiting For EIP Status {0}, time-consuming {1} seconds.".format(status, timeout))
                time.sleep(interval)
            return None
        except Exception as e:
            raise Exception("Waiting For EIP Status {0} Error: {1}.".format(status, e))

    def get_vswitch_status(self, vpc_id, zone_id=None, vswitch_id=None, pagenumber=None, pagesize=None):
        """
        List VSwitches of VPC with their status
        :type vpc_id: string
        :param vpc_id: ID of Vpc from which VSwitch belongs
        :type zone_id: string
        :param zone_id: ID of the Zone
        :type vswitch_id: string
        :param vswitch_id: The ID of the VSwitch to be queried
        :type pagenumber: integer
        :param pagenumber: Page number of the instance status list. The start value is 1. The default value is 1
        :type pagesize: integer
        :param pagesize: The number of lines per page set for paging query. The maximum value is 50 and default
        value is 10
        :return: Returns list of vswitches in VPC with their status
        """
        params = {}
        results = []

        self.build_list_params(params, vpc_id, 'VpcId')
        if zone_id:
            self.build_list_params(params, zone_id, 'ZoneId')
        if vswitch_id:
            self.build_list_params(params, vswitch_id, 'VSwitchId')
        if pagenumber:
            self.build_list_params(params, pagenumber, 'PageNumber')
        if pagesize:
            self.build_list_params(params, pagesize, 'PageSize')

        try:
            results = self.get_status('DescribeVSwitches', params)
        except Exception as ex:
            error_code = ex.error_code
            error_msg = ex.message
            results.append({"Error Code": error_code, "Error Message": error_msg})

        return False, results

    def wait_for_vpc_status(self, vpc_id, status, delay=DefaultWaitForInterval, timeout=DefaultTimeOut):

        try:
            while True:
                vpc = self.describe_vpc_attribute(vpc_id=vpc_id)
                if vpc and str(vpc.status) in [status, str(status).lower()]:
                    return True

                timeout -= delay

                if timeout <= 0:
                    raise Exception("Timeout Error: Waiting for VPC status is %s, time-consuming %d seconds." % (status, timeout))

                time.sleep(delay)
        except Exception as e:
            raise e

    def wait_for_vswitch_status(self, vswitch_id, status, delay=DefaultWaitForInterval, timeout=DefaultTimeOut):
        try:
            while True:
                vsw = self.describe_vswitch_attribute(vswitch_id=vswitch_id)
                if vsw and str(vsw.status) in [status, str(status).lower()]:
                    return True

                timeout -= delay

                if timeout <= 0:
                    raise Exception("Timeout Error: Waiting for VSwitch status is %s, time-consuming %d seconds." % (status, timeout))

                time.sleep(delay)
        except Exception as e:
            raise e

    def wait_for_route_entry_status(self, route_table_id, destination_cidrblock, status, delay=DefaultWaitForInterval, timeout=DefaultTimeOut):
        try:
            tm = 0
            while tm < timeout:
                route_entry = self.get_route_entry_attribute(route_table_id, destination_cidrblock)
                if route_entry and str.lower(route_entry.status) == str.lower(status):
                    return route_entry

                tm += delay

                if tm >= timeout:
                    raise Exception("Timeout Error: Waiting for route entry status is %s, time-consuming %d seconds." % (status, timeout))

                time.sleep(delay)
            return None
        except Exception as e:
            raise Exception("Waiting For route entry Status {0} Error: {1}.".format(status, e))
